/***********************************************************
*                      pC/LZFG 1.20a                       *
*                a FREE - LZFG decompressor                *
*                                                          *
*          by  Frank Goetze  -  www.embedded-os.de         *
************************************************************
*                        pC_LZFG.C                         *
************************************************************
* Revision-Info:                                           *
*                                                          *
* 1.02b     optimize code                                  *
* 1.10a     complete update to U08..U32 types.             *
*           better portability for RTOS-less use impl.     *
* 1.10b     fix an problem of un-init of "t"               *
* 1.20a     complete redesign of "org-lenght" &            *
*           "org-crc" handling (size,endianess,alignment)  *
*           to declare this, the signature is changed      *
*           (in COP too)                                   *
*                                                          *
***********************************************************/

#include "../inc/pC_LZFG.h"

/**********************************************************
  INPUT file-header:    sign + org-len + org-crc + DATA
  
  new since 1.20a:     org-len   4bytes big-endian 
                       org-crc   2bytes big-endian
                       both are not 32bit aligned
**********************************************************/

/*
***********************************************************
*                    FILE-HEADER STRUCTURE
***********************************************************
*/
U08 OS_CONST OS_ROMDATA pC_LZFG[]= {"LZFG\x17\x57"};

/*
************************************************************
*                        CRC-CHECKING
************************************************************
*/
U08 LZFGcrc(U08 OS_HUGE *ldata, U32 lenght, U16 crci) KC51_REENTRANT
{
    U08  i;
    U16  crc;
    U32  byte;

    byte = 0;
    crc  = 0;
    while(byte < lenght) {
        crc ^= ((U16)(ldata[byte++]) << 8);
        for(i=0; i<8; i++) {
            if(crc & 0x8000) {
                crc = (U16)((crc << 1) ^ 0x1021);
            } else {
                crc = (U16)(crc << 1);
            }
        }
    }
    if(crc != crci) return(1);
    return(0);
}


/*
************************************************************
*                     MINI STRING-COMPARE
************************************************************
*/
U08 LZFGstrcmp(U08 OS_HUGE *str0, U08 OS_HUGE *str1, U16 strlen) KC51_REENTRANT
{
    U16  index;

    index = 0;
    while(index < strlen) {
        if(str0[index] != str1[index]) return(1);
        index++;
    }
    return(0);
}


/*
************************************************************
*                   uC/LZFG DECOMPRESSOR
************************************************************
*/
U08 LZFG_Decomp(U08 OS_HUGE *source, U32 source_l, U08 OS_HUGE **dest_pp, U32 *dest_l) KC51_REENTRANT
{
    U08          returnOk, bf;
    U08 OS_HUGE *LZFG_Win;
    U08 OS_HUGE *Decdata_p;
    S16          len, off, t;
    U16          crci;
    U32          ibyte, obyte, cbytes;

    obyte = 0;                                             // start output-bytes
    if(LZFGstrcmp((U08 OS_HUGE *)pC_LZFG, source, sizeof(pC_LZFG)-1) != 0) {
        return(LZFG_noLZFG);                               // control signature
    } else {
        ibyte  = (U32)(sizeof(pC_LZFG) - 1);               // over signature
        cbytes = ((U32)(source[ibyte  ]) << 24) |          // get org-length
                 ((U32)(source[ibyte+1]) << 16) |          // - as U32, big-endian, not aligned
                 ((U32)(source[ibyte+2]) <<  8) |
                 ((U32)(source[ibyte+3])      );
        crci   = ((U16)(source[ibyte+4]) <<  8) |          // get org-crc
                 ((U16)(source[ibyte+5])      );           // - as U16, big-endian, not aligned
        ibyte += 6;                                        // over org-length & org-crc
#ifdef _pC_OS_
        returnOk  = OS_MemAlloc((void OS_HUGE *)(&Decdata_p), cbytes);
#else
        Decdata_p = MEM_ALLOC_D(cbytes);
        if(Decdata_p == NULL) {
            returnOk = LZFG_memERR;
        }
#endif
        if(returnOk != OS_SUCCESS) return(LZFG_memERR);    // malloc of decomp-buffer fails
#ifdef _pC_OS_
        returnOk = OS_MemAlloc((void OS_HUGE *)&LZFG_Win, LZFG_W);
#else
        LZFG_Win = MEM_ALLOC_W(LZFG_W);
        if(Decdata_p == NULL) {
            returnOk = LZFG_memERR;
        }
#endif
        if(returnOk != OS_SUCCESS) {                       // malloc of window fails
#ifdef _pC_OS_
            OS_MemFree((OS_MEM OS_HUGE *)Decdata_p);
#else
            MEM_FREE(Decdata_p);
#endif
            return(LZFG_memERR);
        }
        for(t=0; t<LZFG_W; t++) {                          // init window
            LZFG_Win[t] = 0;
        }
        t = 0;
/***********************************************************
*   for filled-up datastreams (X-Modem) write...           *
*   while(ibyte<source_l && obyte<cbytes) {       ...or... *
***********************************************************/
        while(ibyte < source_l) {                          // up to end of compressed file
            bf = source[ibyte++];
            if(bf & 0xF0) {
                len = 1 + (S16)(bf >> 4);
                off = t - (((S16)(bf & 0x0F) << 8) + (S16)(source[ibyte++])) - 1;
                while(len--) {
                    off &= (LZFG_W - 1);
                    LZFG_Win[t++]      = LZFG_Win[off];
                    Decdata_p[obyte++] = LZFG_Win[off++];
                    t   &= (LZFG_W - 1);
                }
            } else {
                len = 1 + bf;
                while(len--) {
                    LZFG_Win[t++]      = source[ibyte];
                    Decdata_p[obyte++] = source[ibyte++];
                    t   &= (LZFG_W - 1);
                }
            }
        }
#ifdef _pC_OS_
        OS_MemFree((OS_MEM OS_HUGE *)LZFG_Win);
#else
        MEM_FREE(LZFG_Win);
#endif
        returnOk = LZFG_SUCCESS;
        if(obyte != cbytes) {                              // check decompressed-length
            returnOk = LZFG_lenERR;
        } else {
            if(LZFGcrc(Decdata_p, obyte, crci)) {          // check decompressed-crc
                returnOk = LZFG_crcERR;
            }
        }
        if(returnOk != LZFG_SUCCESS) {                     // decompressed with errors...
#ifdef _pC_OS_
            OS_MemFree((OS_MEM OS_HUGE *)Decdata_p);
#else
            MEM_FREE(Decdata_p);
#endif
            return(returnOk);                              // free destination & report error
        }
        *dest_pp = Decdata_p;
        *dest_l  = obyte;
        return(LZFG_SUCCESS);                              // no error - return decompressed data & length
    }
}


